// Copyright (c) 2000
// Utrecht University (The Netherlands),
// ETH Zurich (Switzerland),
// INRIA Sophia-Antipolis (France),
// Max-Planck-Institute Saarbruecken (Germany),
// and Tel-Aviv University (Israel).  All rights reserved.
//
// This file is part of CGAL (www.cgal.org)
//
// $URL: https://github.com/CGAL/cgal/blob/v6.0.1/Cartesian_kernel/include/CGAL/Cartesian/basic_constructions_2.h $
// $Id: include/CGAL/Cartesian/basic_constructions_2.h 50cfbde3b84 $
// SPDX-License-Identifier: LGPL-3.0-or-later OR LicenseRef-Commercial
//
//
// Author(s)     : Herve Bronnimann

#ifndef CGAL_CARTESIAN_BASIC_CONSTRUCTIONS_2_H
#define CGAL_CARTESIAN_BASIC_CONSTRUCTIONS_2_H

#include <CGAL/Cartesian/point_constructions_2.h>
#include <CGAL/Cartesian/line_constructions_2.h>
#include <CGAL/Cartesian/ft_constructions_2.h>

#endif // CGAL_CARTESIAN_BASIC_CONSTRUCTIONS_2_H
