// Copyright (c) 2014  GeometryFactory (France).  All rights reserved.
//
// This file is part of CGAL (www.cgal.org)
//
// $URL: https://github.com/CGAL/cgal/blob/v6.0.1/BGL/include/CGAL/boost/graph/properties_TriMesh_ArrayKernelT.h $
// $Id: include/CGAL/boost/graph/properties_TriMesh_ArrayKernelT.h 50cfbde3b84 $
// SPDX-License-Identifier: LGPL-3.0-or-later OR LicenseRef-Commercial
//
// Author(s)     : Philipp Möller


#ifndef CGAL_PROPERTIES_TRIMESH_ARRAYKERNELT_H
#define CGAL_PROPERTIES_TRIMESH_ARRAYKERNELT_H

#include <OpenMesh/Core/Mesh/TriMesh_ArrayKernelT.hh>
#include <OpenMesh/Core/Mesh/PolyMesh_ArrayKernelT.hh>
#define OPEN_MESH_CLASS OpenMesh::TriMesh_ArrayKernelT<K>
#include <CGAL/boost/graph/properties_OpenMesh.h>

#endif /* CGAL_PROPERTIES_TRIMESH_ARRAYKERNELT_H */
