// Copyright (c) 1998
// Utrecht University (The Netherlands),
// ETH Zurich (Switzerland),
// INRIA Sophia-Antipolis (France),
// Max-Planck-Institute Saarbruecken (Germany),
// and Tel-Aviv University (Israel).  All rights reserved.
//
// This file is part of CGAL (www.cgal.org)
//
// $URL: https://github.com/CGAL/cgal/blob/v6.0.1/Distance_3/include/CGAL/squared_distance_3.h $
// $Id: include/CGAL/squared_distance_3.h 50cfbde3b84 $
// SPDX-License-Identifier: LGPL-3.0-or-later OR LicenseRef-Commercial
//
//
// Author(s)     : Geert-Jan Giezeman


#ifndef CGAL_DISTANCE_3_H
#define CGAL_DISTANCE_3_H

#include <CGAL/Distance_3/Point_3_Point_3.h>
#include <CGAL/Distance_3/Point_3_Weighted_point_3.h>
#include <CGAL/Distance_3/Point_3_Segment_3.h>
#include <CGAL/Distance_3/Point_3_Ray_3.h>
#include <CGAL/Distance_3/Point_3_Line_3.h>
#include <CGAL/Distance_3/Point_3_Triangle_3.h>
#include <CGAL/Distance_3/Point_3_Plane_3.h>
#include <CGAL/Distance_3/Point_3_Tetrahedron_3.h>

#include <CGAL/Distance_3/Weighted_point_3_Weighted_point_3.h>

#include <CGAL/Distance_3/Segment_3_Segment_3.h>
#include <CGAL/Distance_3/Segment_3_Ray_3.h>
#include <CGAL/Distance_3/Segment_3_Line_3.h>
#include <CGAL/Distance_3/Segment_3_Plane_3.h>

#include <CGAL/Distance_3/Ray_3_Ray_3.h>
#include <CGAL/Distance_3/Ray_3_Line_3.h>
#include <CGAL/Distance_3/Ray_3_Plane_3.h>

#include <CGAL/Distance_3/Line_3_Line_3.h>
#include <CGAL/Distance_3/Line_3_Plane_3.h>

#include <CGAL/Distance_3/Triangle_3_Triangle_3.h>

#include <CGAL/Distance_3/Plane_3_Plane_3.h>

#endif // CGAL_DISTANCE_3_H
