// Copyright (c) 1999-2003  ETH Zurich (Switzerland).
// All rights reserved.
//
// This file is part of CGAL (www.cgal.org).
//
// $URL: https://github.com/CGAL/cgal/blob/v6.0.1/Bounding_volumes/include/CGAL/minimum_enclosing_quadrilateral_2.h $
// $Id: include/CGAL/minimum_enclosing_quadrilateral_2.h 50cfbde3b84 $
// SPDX-License-Identifier: GPL-3.0-or-later OR LicenseRef-Commercial
//
//
// Author(s)     : Michael Hoffmann <hoffmann@inf.ethz.ch> and
//                 Emo Welzl <emo@inf.ethz.ch>

#ifndef CGAL_MINIMUM_ENCLOSING_QUADRILATERAL_2_H
#define CGAL_MINIMUM_ENCLOSING_QUADRILATERAL_2_H 1

#include <CGAL/license/Bounding_volumes.h>

#define CGAL_DEPRECATED_HEADER "<CGAL/minimum_enclosing_quadrilateral_2.h>"
#define CGAL_REPLACEMENT_HEADER "<CGAL/min_quadrilateral_2.h>"
#include <CGAL/Installation/internal/deprecation_warning.h>

#include <CGAL/min_quadrilateral_2.h>

#endif // ! (CGAL_MINIMUM_ENCLOSING_QUADRILATERAL_2_H)

// ----------------------------------------------------------------------------
// ** EOF
// ----------------------------------------------------------------------------
