// Copyright (c) 2006 Foundation for Research and Technology-Hellas (Greece).
// All rights reserved.
//
// This file is part of CGAL (www.cgal.org).
//
// $URL: https://github.com/CGAL/cgal/blob/v6.0.1/Voronoi_diagram_2/include/CGAL/Voronoi_diagram_2/basic.h $
// $Id: include/CGAL/Voronoi_diagram_2/basic.h 50cfbde3b84 $
// SPDX-License-Identifier: GPL-3.0-or-later OR LicenseRef-Commercial
//
//
// Author(s)     : Menelaos Karavelas <mkaravel@iacm.forth.gr>

#ifndef CGAL_VORONOI_DIAGRAM_2_BASIC_H
#define CGAL_VORONOI_DIAGRAM_2_BASIC_H 1

#include <CGAL/license/Voronoi_diagram_2.h>


#include <CGAL/basic.h>
#include <cstdlib>

#define CGAL_VORONOI_DIAGRAM_2_INS CGAL::VoronoiDiagram_2::Internal

#endif // CGAL_VORONOI_DIAGRAM_2_BASIC_H
