// Copyright (c) 2007  GeometryFactory (France).  All rights reserved.
//
// This file is part of CGAL (www.cgal.org)
//
// $URL: https://github.com/CGAL/cgal/blob/v6.0.1/HalfedgeDS/include/CGAL/boost/graph/properties_HalfedgeDS_default.h $
// $Id: include/CGAL/boost/graph/properties_HalfedgeDS_default.h 50cfbde3b84 $
// SPDX-License-Identifier: LGPL-3.0-or-later OR LicenseRef-Commercial
//
//
// Author(s)     : Andreas Fabri, Fernando Cacciola

#ifndef CGAL_BOOST_GRAPH_PROPERTIES_HALFEDGEDS_DEFAULT_H
#define CGAL_BOOST_GRAPH_PROPERTIES_HALFEDGEDS_DEFAULT_H

#define CGAL_HDS_TMPLT Gt, class I,  class A
#define CGAL_HDS_CLASS CGAL::HalfedgeDS_default<Gt,I,A>

#include <CGAL/boost/graph/properties_HalfedgeDS_base.h>

#endif // CGAL_BOOST_GRAPH_PROPERTIES_HALFEDGEDS_DEFAULT_H
