// Copyright (c) 2016  GeometryFactory Sarl (France)
//
// This file is part of CGAL (www.cgal.org)
//
// $URL: https://github.com/CGAL/cgal/blob/v6.0.1/Installation/include/CGAL/license.h $
// $Id: include/CGAL/license.h 50cfbde3b84 $
// SPDX-License-Identifier: LGPL-3.0-or-later OR LicenseRef-Commercial
//
//
// Author     : Andreas Fabri

#ifndef CGAL_LICENSE_CHECK_H
#define CGAL_LICENSE_CHECK_H

//#define CGAL_LICENSE_WARNING  1   // in order to get a warning during compilation
//#define CGAL_LICENSE_ERROR    1   // in order to get an error during compilation

// if no such define exists, the package is used under the GPL or LGPL
// Otherwise the number encodes the date year/month/day
// Any release made b......
// e.g.
//#define CGAL_AABB_TREE_COMMERCIAL_LICENSE  20140101

#endif // CGAL_LICENSE_CHECK_H
