//
// Copyright (c) 2017 GeometryFactory (France).  All Rights Reserved.
//
// This file is part of CGAL (www.cgal.org)
//
// $URL: https://github.com/CGAL/cgal/blob/v6.0.1/Subdivision_method_3/include/CGAL/subdivision_method_3.h $
// $Id: include/CGAL/subdivision_method_3.h 50cfbde3b84 $
// SPDX-License-Identifier: LGPL-3.0-or-later OR LicenseRef-Commercial
//
//
// Author(s):
//

/**
* \ingroup PkgSurfaceSubdivisionMethod3Ref
* \file CGAL/subdivision_method_3.h
* Convenience header file including the headers for all
* the free functions of this package.
*/

#include <CGAL/Subdivision_method_3/subdivision_methods_3.h>
