// Copyright (c) 1997-2000
// Utrecht University (The Netherlands),
// ETH Zurich (Switzerland),
// INRIA Sophia-Antipolis (France),
// Max-Planck-Institute Saarbruecken (Germany),
// and Tel-Aviv University (Israel).  All rights reserved.
//
// This file is part of CGAL (www.cgal.org)
//
// $URL: https://github.com/CGAL/cgal/blob/v6.0.1/Kernel_d/include/CGAL/Kernel_d/Interface_classes.h $
// $Id: include/CGAL/Kernel_d/Interface_classes.h 50cfbde3b84 $
// SPDX-License-Identifier: LGPL-3.0-or-later OR LicenseRef-Commercial
//
//
// Author(s)     : Michael Seel <seel@mpi-sb.mpg.de>

#ifndef CGAL_INTERFACE_CLASSES_H
#define CGAL_INTERFACE_CLASSES_H

namespace CGAL {

template <class R> class Point_d;
template <class R> class Vector_d;
template <class R> class Direction_d;
template <class R> class Hyperplane_d;
template <class R> class Sphere_d;
template <class R> class Iso_box_d;
template <class R> class Segment_d;
template <class R> class Ray_d;
template <class R> class Line_d;
template <class R> class Aff_transformation_d;

} //namespace CGAL

#endif //CGAL_INTERFACE_CLASSES_H
