// Copyright (C) 2018  GeometryFactory Sarl
//
// This file is part of CGAL (www.cgal.org)
//
// $URL: https://github.com/CGAL/cgal/blob/v6.0.1/Installation/include/CGAL/Kernel_traits_fwd.h $
// $Id: include/CGAL/Kernel_traits_fwd.h 50cfbde3b84 $
// SPDX-License-Identifier: LGPL-3.0-or-later OR LicenseRef-Commercial
//

#ifndef CGAL_KERNEL_TRAITS_FWD_H
#define CGAL_KERNEL_TRAITS_FWD_H

/// \file Kernel_traits_fwd.h
/// Forward declarations of the `Kernel_traits` class.

#ifndef DOXYGEN_RUNNING
namespace CGAL {

template <class T> struct Kernel_traits;

} // CGAL
#endif

#endif // CGAL_KERNEL_TRAITS_FWD_H
