// Copyright (c) 2014 GeometryFactory
// All rights reserved.
//
// This file is part of CGAL (www.cgal.org).
//
// $URL: https://github.com/CGAL/cgal/blob/v6.0.1/Surface_mesh_shortest_path/include/CGAL/Surface_mesh_shortest_path.h $
// $Id: include/CGAL/Surface_mesh_shortest_path.h 50cfbde3b84 $
// SPDX-License-Identifier: GPL-3.0-or-later OR LicenseRef-Commercial
//
// Author(s)     : Stephen Kiazyk

#ifndef CGAL_SURFACE_MESH_SHORTEST_PATH_H
#define CGAL_SURFACE_MESH_SHORTEST_PATH_H

#include <CGAL/license/Surface_mesh_shortest_path.h>


/**
 * \ingroup PkgSurfaceMeshShortestPathRef
 * \file CGAL/Surface_mesh_shortest_path.h
 * Convenience header file only including `CGAL/Surface_mesh_shortest_path/Surface_mesh_shortest_path.h`
 * and `CGAL/Surface_mesh_shortest_path/Surface_mesh_shortest_path_traits.h`.
 */

#include <CGAL/Surface_mesh_shortest_path/Surface_mesh_shortest_path.h>
#include <CGAL/Surface_mesh_shortest_path/Surface_mesh_shortest_path_traits.h>

#endif // CGAL_SURFACE_MESH_SHORTEST_PATH_H
