// Copyright (c) 2011 GeoemtryFactory
// All rights reserved.
//
// This file is part of CGAL (www.cgal.org)
//
// $URL: https://github.com/CGAL/cgal/blob/v6.0.1/Number_types/include/CGAL/Sqrt_extension_fwd.h $
// $Id: include/CGAL/Sqrt_extension_fwd.h 50cfbde3b84 $
// SPDX-License-Identifier: LGPL-3.0-or-later OR LicenseRef-Commercial
//
//
// Author(s)     : Sebastien Loriot <Sebastien.Loriot@geometryfactory.com>
//



#ifndef CGAL_SQRT_EXTENSION_FWD_H
#define CGAL_SQRT_EXTENSION_FWD_H

#include <CGAL/tags.h>

namespace CGAL{

template <class NT_,class ROOT_, class AllowCompareDifferentExtensions = ::CGAL::Tag_false, class FilterPredicates = ::CGAL::Tag_false >
class Sqrt_extension;

} //namespace CGAL

#endif // CGAL_SQRT_EXTENSION_FWD_H
