// Copyright (c) 2007  GeometryFactory (France).  All rights reserved.
//
// This file is part of CGAL (www.cgal.org)
//
// $URL: https://github.com/CGAL/cgal/blob/v6.0.1/Polyhedron/include/CGAL/boost/graph/halfedge_graph_traits_Polyhedron_3.h $
// $Id: include/CGAL/boost/graph/halfedge_graph_traits_Polyhedron_3.h 50cfbde3b84 $
// SPDX-License-Identifier: GPL-3.0-or-later OR LicenseRef-Commercial
//
//
// Author(s)     : Andreas Fabri, Fernando Cacciola


#ifndef CGAL_BOOST_GRAPH_HALFEDGE_GRAPH_TRAITS_POLYHEDRON_3_H
#define CGAL_BOOST_GRAPH_HALFEDGE_GRAPH_TRAITS_POLYHEDRON_3_H

#include <CGAL/license/Polyhedron.h>


#define CGAL_DEPRECATED_HEADER "<CGAL/boost/graph/halfedge_graph_traits_Polyhedron_3.h>"
#define CGAL_REPLACEMENT_HEADER "<CGAL/boost/graph/graph_traits_Polyhedron_3.h>"
#include <CGAL/Installation/internal/deprecation_warning.h>

#include <CGAL/boost/graph/graph_traits_Polyhedron_3.h>

#endif // CGAL_BOOST_GRAPH_HALFEDGE_GRAPH_TRAITS_POLYHEDRON_3_H
